#ifndef pan3d_create_h
#define pan3d_create_h
////
//
/////////////////////////////////////////////////////////////////////////////////////////
//																													//
//		PAN3D_CREATE	-	Function for turning a panel into a true 3D map-entity			//
//																													//
/////////////////////////////////////////////////////////////////////////////////////////
//																													//
//		Author	:	EvilSOB																					//
//		Written	: 	30-07-2010																				//
//		Version	:	2.5																						//
//																													//
//		Requirements:	Acknex.h		(only tested with A7.86.0 so far								//
//																													//
/////////////////////////////////////////////////////////////////////////////////////////
//																													//
//		pan3d_create(PANEL* pan, vector* pos, void* act, var mode);								//
//=====================================================================================//
//		Create a new entity based of the visual output of "panel".  (using target_map)	//
//		!!!ACTIVE panel controls (buttons and sliders) will not react to mouse, YET!!!	//
//		"panel" is no longer visible once its attached to the new entity.						//
//		The function returns a pointer to the new entity. This entity is then used for 	//
//		all visual effect and position controls. (including shaders)			 				//
//		If ENTITY's SHOW flag is turned off, panel-entity is merely "hidden".				//
//		BUT, if the PANEL's SHOW flag is turned off, both PANEL and ENTITY are destroyed.//
//-------------------------------------------------------------------------------------//
//		Parameters:																								//
//			panel	 - pointer to the "source" panel. 							  (required)		//
//			pos	 - 3D position of entity placement.							  (see POS notes)	//
//			act 	 - optional user action to apply to new entity.			  (see ACT notes) //
//			mode   - optional flags to setup integral action properties.  (see MODE notes)//
//-------------------------------------------------------------------------------------//
//		Returns:																									//
//			Pointer to panel-entity.								 										//
//																													//
//=====================================================================================//
//		NOTES:																									//
//-------------------------------------------------------------------------------------//
//		MODE : Most of the following mode flags can be combined.									//
//			ACT_NONE 		means NO integral position/rotation actions are performed.		//
//			FOLLOW_ME		content of 'me' at the time of calling is the follow-target.	//
//			FOLLOW_YOU		content of 'you' at the time of calling is the follow-target.	//
//			FOLLOW_CAMERA	the view "camera" is the follow-target for position & rotation.// 
//			ORBIT_FOLLOW	panel keeps its position relative to parents rotation.			//
//			FACE_CAMERA		Sprite always faces camera.												//
//			FACE_FLAT		Sprite always appears "flat". 	(perpendicular to screen)		//
//																													//
//-------------------------------------------------------------------------------------//
//		POS : 																									//
//			Where MODE sets a follow-target, the parameter POS is used as an offset 		//
//			position from the parent, otherwise it is a global position.						//
//			------------------------------------------------------------------------------//
//			When a follow-target has been set, and the ORBIT_FOLLOW flag is set, then the //
//			position of the panel stays relative with the rotation of the parent as it 	//
//			was at the time of the panels creation, including the position-offset.			//
//																													//
//-------------------------------------------------------------------------------------//
//		ACT : User defined actions																			//
//			This action is attached to the panel using all the same requirements and		//
//			limitations as any normal entity action.													//
//																													//
//-------------------------------------------------------------------------------------//
//																													//
//																													//
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	skill numbers may be changed for your project. These MAY conflict with A7 templates	//
#define	ep_p_off			skill90	// target X/Y/Z offset 				(if FOLLOW or ORBIT)	//
#define	ep_p_rot			skill93	// target pan/tilt/roll offset	(if FOLLOW or ORBIT)	//
#define	ep_panel			skill96	// void* to source panel 										//
#define	ep_size_x		skill97	// 3D-panel dimensions in pixels 	[READ-ONLY]			//
#define	ep_size_y		skill98	// 	size is based on panel size_x/y * scale_x/y		//
#define	ep_mode			skill99	// Integral Action flags. See below							//
/////////////////////////////////////////////////////////////////////////////////////////
#define	ep_parent		parent	// FOLLOW target 						(if FOLLOW enabled)	//
#define	ACT_NONE			0			// NO integral actions.											//
#define	FOLLOW_ME		1			// Make current "me"  the follow-target.					//
#define	FOLLOW_YOU		2			// Make current "you" the follow-target.					//
#define	FOLLOW_CAMERA	4			// Make "camera" the follow-target.							// 
#define	ORBIT_FOLLOW	8			// 'orbit' with follow-target as it rotates.				//
#define	FACE_CAMERA		16			//	Sprite always faces camera									//
#define	FACE_FLAT		32			//	Sprite always "flat". 	(perpendicular to screen)	//
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//																													//
//																													//
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
void	pan3d_worker(ENTITY* ME, PANEL* pan);		//prototype for worker thread				//
#include <d3d9.h>																								//
LPD3DXMESH *panel_mesh;																						//
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
ENTITY* pan3d_create(PANEL* pan, VECTOR* pos, void* act, var mode)							//
{	if(!pan)		return(NULL);		//invalid without a panel										//
	ENTITY* ME = ent_create(CUBE_MDL, nullvector, act);		ent_clone(ME);					//
	ent_setmesh(ME, panel_mesh, 0,0);								ent_clone(ME);					//
	//////////////////////////////////////////////////////////////////////////////////////
	ME.ep_mode = mode;	ME.ep_panel=(void*)pan;		if(pos)	vec_set(ME.x, pos);			//
	if(mode&FOLLOW_ME)	ME.ep_parent=me;	if(mode&FOLLOW_YOU)	ME.ep_parent=you;			//
	pan3d_worker(ME, pan);		return(ME);													 }			//
/////////////////////////////////////////////////////////////////////////////////////////
																													//
typedef struct BUTTON																						//
{	struct	BUTTON *next;																					//
				long	num,x,y;         																		//
				DWORD color;																					//
				long	dx,dy;																					//
				BMAP	*MapOn, *MapOff, *MapOver, *MapOverOff;										//
				void	*actionOn, *actionOff, *actionOver; 											//
				int	status, type;																			//
} BUTTON;																										//
//																													//
#define TYPE_TOGGLE		2																					//
#define TYPE_RADIO		4																					//
//																													//
BUTTON* button_next(PANEL* panel, BUTTON* buttonn)													//
{	//returns next button from "buton" for this panel 												//
	//if 'buttonn' is NULL, then returns first button												//
	if(!panel)	return(NULL);	//no panel to work from												//
	if(!buttonn)	return(*((long*)&panel.pad[8]));													//
	return(*((long*)buttonn));								 }												//
//																													//
void	_ent_panel_funct_call(var butt, PANEL* pan);	//for calling triggered functions	//
//																													//
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//																													//
void exit_event()	{	me=NULL;		while(me=ent_next(me))	ent_setmesh(me,0,0,0);	}		//
//																													//
void	make_panelmesh_startup()																			//
{	//////////////////////////////////////////////////////////////////////////////////////
	//	static hard-coded mesh		////																	//
	static short Ibuff[12] 	={	0,2,1,0,3,2,3,1,2,3,0,1 };											//
	static byte Vbuff[224]	={ 0,128,95,56,0,0,200,66,47,0,200,66,0,0,128,191,0,0,0,128,//
		28,10,15,53,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	//
		128,95,56,0,0,200,194,47,0,200,66,0,0,128,63,0,0,0,128,28,10,15,181,0,0,0,0,0,0,	//
		128,63,0,0,0,0,0,0,128,63,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,95,184,0,0,200,	//
		194,47,0,200,194,0,0,128,191,0,0,0,128,28,10,15,53,0,0,128,63,0,0,128,63,0,0,128,//
		63,0,0,128,63,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,95,184,0,0,200,66,47,0,200,	//
		194,0,0,128,63,0,0,0,128,28,10,15,181,0,0,128,63,0,0,0,0,0,0,128,63,0,0,0,0,0,0,	//
		0,0,0,0,0,0,0,0,0,0,0,0,0,0	};																	//
	//////////////////////////////////////////////////////////////////////////////////////
	on_exit = exit_event;																					//
	LPDIRECT3DDEVICE9 pd3d_dev = pd3ddev;		LPD3DXMESH pMesh;									//
	D3DVERTEX* pVertBuff;							short* pFaceBuff;									//
	//////////////////////////////////////////////////////////////////////////////////////
	if(SUCCEEDED(D3DXCreateMeshFVF(0x04,0x04,0x220,D3DFVF_D3DVERTEX,pd3d_dev,&pMesh)))	//
	{	if(SUCCEEDED(pMesh->LockVertexBuffer(0x0, &pVertBuff)))									//
		{	memcpy(pVertBuff, Vbuff, sizeof(D3DVERTEX)*0x4);										//
			pMesh->UnlockVertexBuffer();								}									//
		if(SUCCEEDED(pMesh->LockIndexBuffer(0x0, &pFaceBuff)))									//
		{	memcpy(pFaceBuff, Ibuff, (sizeof(short)*0x3)*0x4);										//
			pMesh->UnlockIndexBuffer();	  						}	}									//
		panel_mesh = pMesh;												}									//
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
void	pan3d_worker(ENTITY* ME, PANEL* pan)															//
{	wait(1);		if(!(me=ME))		return;		//somehow invalid entity						//
	my.ep_size_x = pan.size_x*pan.scale_x;		my.ep_size_y = pan.size_y*pan.scale_y;		//
	ent_setskin(me, (my.lightmap=bmap_createblack(my.ep_size_x, my.ep_size_y,32)), 1);	//
	if(!is(pan,TRANSLUCENT))	pan.alpha = 100;														//			
	pan.pos_x=pan.pos_y = 0;	my.alpha=pan.alpha;		set(my,TRANSLUCENT);					//
	proc_mode  = PROC_GLOBAL|PROC_LATE;				pan.target_map=my.lightmap;				//
	VECTOR tmpV; 		vec_zero(my.ep_p_off);		vec_zero(my.ep_p_rot);						//
	if(my.ep_parent)	{	vec_set(my.ep_p_off, my.x);												//
		if(my.ep_mode&ORBIT_FOLLOW)		vec_set(my.ep_p_rot, my.ep_parent.pan);	}		//
	else if(my.ep_mode&FOLLOW_CAMERA)																	//
	{	vec_set(my.ep_p_off, my.x);		vec_set(my.ep_p_rot, camera.pan);			}		//
	VECTOR fromV,toV;	CONTACT* c = ent_getvertex(me,0,1);			ent_setvertex(me,c,1);	//
	D3DVERTEX* vBuff = c.v;							vBuff[0].y=vBuff[3].y= my.ep_size_y/2;		//
	vBuff[1].y=vBuff[2].y=-my.ep_size_y/2;		vBuff[0].z=vBuff[1].z= my.ep_size_x/2;		//
	vBuff[2].z=vBuff[3].z=-my.ep_size_x/2;		c_setminmax(me);									//
	BUTTON *butt;		long status,butt_hit,butt_over=0;		BMAP* butt_over_bmap=NULL;	//
	//////////////////////////////////////////////////////////////////////////////////////
	while(me && is(pan,SHOW))																				//
	{	//------------------------------------------------------------							//
		// perform "integral action"s of POSITION & ROTATION types								//
		if(my.ep_mode&FOLLOW_CAMERA)	//FOLLOW camera												//
		{	if(my.ep_mode&ORBIT_FOLLOW)		//'orbit' as FOLLOW_ target rotates				//
			{	vec_rotate(vec_set(tmpV, my.ep_p_off), camera.pan);								//
				vec_set(my.x, vec_add(tmpV, camera.x));								}				//
			else				//else Just FOLLOW at set absolute distance							//
				vec_set(my.x, vec_add(vec_set(tmpV,camera.x), my.ep_p_off));	}				//
		else if(my.parent)		//FOLLOW_ME or YOU is enabled										//
		{	if(my.ep_mode&ORBIT_FOLLOW)		//'orbit' as FOLLOW_ target rotates				//
			{	vec_rotate(vec_set(tmpV, my.ep_p_off), my.parent.pan);							//
				vec_set(my.x, vec_add(tmpV, my.parent.x));		  					}				//
			else				//else Just FOLLOW at set absolute distance							//
				vec_set(my.x, vec_add(vec_set(tmpV,my.parent.x), my.ep_p_off));}				//
		// perform "integral action"s of pure ROTATION types										//
		if(my.ep_mode&FACE_CAMERA)		//always face camera											//
		{	vec_to_angle(my.pan,vec_sub(vec_set(vector(0,0,0),camera.x),my.x));				//
			ang_rotate(my.pan, vector(180,0,0));										}				//
		else if(my.ep_mode&FACE_FLAT)	//always "perpendicular" to screen						//
		{	vec_set(my.pan, camera.pan);													}				//
		//------------------------------------------------------------							//
		//respond to panel size_x, size_y, scale_x, scale_y if changed							//
		my.ep_size_x=pan.size_x*pan.scale_x;	my.ep_size_y=pan.size_y*pan.scale_y;		//
		if((my.ep_size_x!=my.lightmap.width)||(my.ep_size_x!=my.lightmap.width))			//
		{	bmap_remove(my.lightmap);																		//
			my.lightmap = bmap_createblack(my.ep_size_x, my.ep_size_y,32);						//
			ent_setskin(me, (pan.target_map=my.lightmap), 1);										//
			c = ent_getvertex(me,0,1);			ent_setvertex(me,c,1);		vBuff = c.v;		//
			vBuff[0].y=vBuff[3].y= my.ep_size_y/2;	vBuff[1].y=vBuff[2].y=-my.ep_size_y/2;	//
			vBuff[0].z=vBuff[1].z= my.ep_size_x/2;	vBuff[2].z=vBuff[3].z=-my.ep_size_x/2;	//
			c_setminmax(me);																				}	//
		//------------------------------------------------------------							//
		// perform "integral action"s of CONTROLS-REACT-TO-MOUSE	types							//
		if(mouse_ent==me)																						//
		{	vec_set(toV, vec_set(fromV, mouse_cursor));	fromV.z=0;	toV.z=mouse_range;	//
			vec_for_screen(toV, camera);			vec_for_screen(fromV, camera);				//
			if(c_trace(fromV, toV, USE_POLYGON|SCAN_TEXTURE))										//
			{	//=======	Process Buttons	=============================================//
				butt=NULL;		while((butt=button_next(pan,butt)))									//
				{	//reset MapOver swapping																//
					if(butt_over==(butt.num+1))															//
					{	if(butt_over_bmap)																	//
						{	if(butt.type&TYPE_TOGGLE)														//
							{	if(button_state(pan,butt.num+1,-1))	butt.MapOn=butt_over_bmap;	//
								else						butt.MapOff = butt_over_bmap;	}				//
							else							butt.MapOff = butt_over_bmap;	}				//
						butt_over_bmap = NULL;												}				//
					////	calculate IF this button is in contact										//
					butt_hit  = (hit.u1>butt.x) && (hit.u1<(butt.x+butt.dx));					//
					butt_hit &= (hit.v1>butt.y) && (hit.v1<(butt.y+butt.dy));					//
					//----------------------------------------------------------------------//
					if(butt_hit)		////	if this button IS in contact							//
					{	//activate actionOver ?																// 
						if(!(status&1)&&!mouse_left)														//
						{	status|=1;	if(_ent_panel_funct_call=butt.actionOver)					//
												_ent_panel_funct_call(butt.num+1, pan);	}			//
						//click and actionClick																//
						if(mouse_left&&!(status&2))		{	status|=2;								//
							if(butt.type&TYPE_TOGGLE)														//
								button_state(pan,butt.num+1,abs((butt.type&1)-1));					//
							else if(butt.type&TYPE_RADIO)													//
								button_state(pan,butt.num+1, 1);											//
							if(_ent_panel_funct_call=butt.actionOn)									//
								_ent_panel_funct_call(butt.num+1, pan);				}				//
						//1st half of actionLeave  (release inside button area)					//
						else if(!mouse_left&&(status&2))	{	status&=~2;								//
							if(_ent_panel_funct_call=butt.actionOff)									//
								_ent_panel_funct_call(butt.num+1, pan);				}				//
						//bmapOver/bmapOverOff swapping													//
						if((butt.type&TYPE_TOGGLE)&&(!mouse_left))									//
						{	if(button_state(pan,butt.num+1,-1)&&(butt.MapOn!=butt.MapOver))	//
							{	butt_over=butt.num+1;		butt_over_bmap=butt.MapOn;				//
								butt.MapOn=butt.MapOver;									 }				//
							else 	if(butt.MapOff!=butt.MapOverOff)										//
							{	butt_over=butt.num+1;		butt_over_bmap=butt.MapOff;			//
								butt.MapOff=butt.MapOverOff;							}	  }			//
						else	if((butt.MapOff!=butt.MapOver)&&(!mouse_left))						//
						{	butt_over=butt.num+1;			butt_over_bmap=butt.MapOff;			//
							butt.MapOff=butt.MapOver;									}	  }			//
					//----------------------------------------------------------------------//
					else					////if button NOT in contact									//
					{	//2nd half of functionOff  (leave button area)								//
						if(!mouse_left&&(butt_over==(butt.num+1)))									//
						{	if(_ent_panel_funct_call=butt.actionOff)									//
								_ent_panel_funct_call(butt.num+1, pan);	}							//
						//reset mouse-clicks																	//
						if(butt_over==(butt.num+1))	status&=~1;			}							//
		}	}	}	//======================================================================//
		//------------------------------------------------------------							//
		my.alpha=pan.alpha;		//always inherit transparency										//
		wait(1);																					}				//
	//////////////////////////////////////////////////////////////////////////////////////
	if(!me)	//hide panel because only entity was removed											//
	{	bmap_remove(pan.target_map);	pan.target_map=NULL;											//
		if(is(pan,SHOW))	reset(pan,SHOW);																//
		else					pan_remove(pan);					 }											//
	else		//remove everything because PANEL.SHOW was turned off								//	
	{	bmap_remove(pan.target_map);	pan.target_map=NULL;											//
		reset(pan,SHOW);					pan_remove(pan);												//
		ent_remove(me);											 }											//
}																													//
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//
//
////
#endif	// pan3d_create_h
